@@ -75,7 +75,8 @@ class Ilo(object):
         protocol to use, but you can override this by setting protocol to
         ILO_RAW or ILO_HTTP. Use ILO_LOCAL to avoid using a network connection
         and use hponcfg instead. Username and password are ignored for ILO_LOCAL
-        connections."""
+        connections. Set delayed to True to make python-hpilo not send requests
+        immediately, but group them together. See :func:`call_delayed`"""
 
     XML_HEADER = b('<?xml version="1.0"?>\r\n')
     HTTP_HEADER = "POST /ribcl HTTP/1.1\r\nHost: localhost\r\nContent-Length: %d\r\nConnection: Close%s\r\n\r\n"
@@ -85,15 +86,18 @@ class Ilo(object):
     if platform.system() == 'Windows':
         hponcfg = 'C:\Program Files\HP Lights-Out Configuration Utility\cpqlocfg.exe'
 
-    def __init__(self, hostname, login=None, password=None, timeout=60, port=443, protocol=None):
+    def __init__(self, hostname, login=None, password=None, timeout=60, port=443, protocol=None, delayed=False):
         self.hostname = hostname
-        self.login = login or 'Administrator'
+        self.login    = login or 'Administrator'
         self.password = password or 'Password'
         self.timeout  = timeout
         self.debug    = 0
         self.port     = port
         self.protocol = protocol
         self.cookie   = None
+        self.delayed  = delayed
+        self._elements = None
+        self._processors = []
 
     def __str__(self):
         return "iLO interface of %s" % self.hostname
@@ -336,6 +340,10 @@ class Ilo(object):
             # We must be using iLO2 or older, they don't do HTTP for XML requests
             # This case is only triggered by the protocol detection
             header = None
+
+        elif not data.startswith('<?xml'):
+            raise IloError("Remote returned bogus data, maybe it's not an iLO")
+
         else:
             header = None
 
@@ -345,8 +353,14 @@ class Ilo(object):
 
     def _root_element(self, element, **attrs):
         """Create a basic XML structure for a message. Return root and innermost element"""
-        root = etree.Element('RIBCL', VERSION="2.0")
-        login = etree.SubElement(root, 'LOGIN', USER_LOGIN=self.login, PASSWORD=self.password)
+        if not self.delayed or not self._elements:
+            root = etree.Element('RIBCL', VERSION="2.0")
+            login = etree.SubElement(root, 'LOGIN', USER_LOGIN=self.login, PASSWORD=self.password)
+        if self.delayed:
+            if self._elements:
+                root, login = self._elements
+            else:
+                self._elements = (root, login)
         element = etree.SubElement(login, element, **attrs)
         return root, element
 
@@ -407,25 +421,25 @@ class Ilo(object):
                 # HP is not best friends with consistency. Sometimes there are
                 # attributes, sometimes child tags and sometimes text nodes. Oh
                 # well, deal with it :)
-                if list(elt):
+                if elt.attrib and list(elt):
                     val = self._element_to_dict(elt)
+                elif list(elt):
+                    val = self._element_to_list(elt)
                 elif elt.text:
-                    val = elt.text
+                    val = elt.text.strip()
                 elif elt.attrib:
                     val = self._element_to_dict(elt)
 
             val = self._coerce(val)
 
             if unit:
-                if isinstance(retval, list):
-                    retval.append((val,unit))
-                else:
-                    retval[key] = (val, unit)
+                val = (val, unit)
+            if isinstance(retval, list):
+                retval.append(val)
+            elif key in retval:
+                retval[key].update(val)
             else:
-                if isinstance(retval, list):
-                    retval.append(val)
-                else:
-                    retval[key] = val
+                retval[key] = val
         return retval
 
     def _element_to_dict(self, element):
@@ -433,8 +447,29 @@ class Ilo(object):
         retval = {}
         for key, val in element.attrib.items():
             retval[key.lower()] = self._coerce(val)
+        if list(element):
+            fields = []
+            for child in element.getchildren():
+                if child.tag == 'FIELD':
+                    fields.append(self._element_to_dict(child))
+            if fields:
+                names = [x['name'] for x in fields]
+                if len(names) == len(set(names)):
+                    # Field names are unique, treat them like attributes
+                    for field in fields:
+                        retval[field['name']] = field['value']
+                else:
+                    # Field names are not unique, such as the name "MAC"
+                    retval['fields'] = fields
         return retval
 
+    def _element_to_list(self, element):
+        tagnames = [x.tag for x in element]
+        if len(set(tagnames)) == 1:
+            return [self._element_children_to_dict(x) for x in element]
+        else:
+            return [(child.tag.lower(), self._element_to_dict(child)) for child in element]
+
     def _coerce(self, val):
         """Do some data type coercion: unquote, turn integers into integers and
            Y/N into booleans"""
@@ -448,6 +483,8 @@ class Ilo(object):
         return val
 
     def _raw(self, *tags):
+        if self.delayed:
+            raise IloError("Cannot use raw tags in delayed mode")
         root, inner = self._root_element(tags[0][0], **(tags[0][1]))
         for t in tags[1:]:
             inner = etree.SubElement(inner, t[0], **t[1])
@@ -458,32 +495,40 @@ class Ilo(object):
         fd.close()
         return ret
 
-    def _info_tag(self, infotype, tagname, returntags=None, attrib={}):
+    def _info_tag(self, infotype, tagname, returntags=None, attrib={}, process=lambda x: x):
         root, inner = self._root_element(infotype, MODE='read')
         etree.SubElement(inner, tagname, **attrib)
+        if self.delayed:
+            self._processors.append([self._process_info_tag, returntags or [tagname], process])
+            return
         header, message = self._request(root)
+        return self._process_info_tag(message, returntags or [tagname], process)
 
-        # Use tagname as returntags if returntags isn't specificed
+    def _process_info_tag(self, message, returntags, process):
         if isinstance(returntags, basestring):
             returntags = [returntags]
-        elif returntags is None:
-            returntags = [tagname]
 
         for tag in returntags:
             if message.find(tag) is None:
                 continue
             message = message.find(tag)
             if list(message):
-                return self._element_children_to_dict(message)
+                return process(self._element_children_to_dict(message))
             else:
-                return self._element_to_dict(message)
+                return process(self._element_to_dict(message))
         raise IloError("Expected tag '%s' not found" % "' or '".join(returntags))
 
-    def _info_tag2(self, infotype, tagname, returntag=None, key=None):
+    def _info_tag2(self, infotype, tagname, returntag=None, key=None, process=lambda x: x):
         root, inner = self._root_element(infotype, MODE='read')
         etree.SubElement(inner, tagname)
+        if self.delayed:
+            self._processors.append([self._process_info_tag2, returntag or tagname, process])
+            return
         header, message = self._request(root)
-        message = message.find(returntag or tagname)
+        return self._process_info_tag2(message, returntag or tagname, process)
+
+    def _process_info_tag2(self, message, returntag, process):
+        message = message.find(returntag)
 
         if key:
             retval = {}
@@ -495,7 +540,7 @@ class Ilo(object):
                 retval[elt[key]] = elt
             else:
                 retval.append(elt)
-        return retval
+        return process(retval)
 
     def _control_tag(self, controltype, tagname, returntag=None, attrib={}, elements=[], text=None):
         root, inner = self._root_element(controltype, MODE='write')
@@ -504,6 +549,8 @@ class Ilo(object):
             inner.text = text
         for element in elements:
             inner.append(element)
+        if self.delayed:
+            return
         header, message = self._request(root)
         if message is None:
             return None
@@ -516,6 +563,30 @@ class Ilo(object):
         else:
             return self._element_to_dict(message)
 
+    def call_delayed(self):
+        """In delayed mode, calling a method on an iLO object will not cause an
+           immediate callout to the iLO. Instead, the method and parameters are
+           stored for future calls of this method. This method makes one
+           connection to the iLO and sends all commands as one XML document.
+           This speeds up applications that make many calls to the iLO by
+           removing seconds of overhead per call.
+
+           The return value of call_delayed is a list of return values for
+           individual methods that don't return None. This means that there may
+           be fewer items returned than methods called as only `get_*` methods
+           return data"""
+
+        if not self._elements:
+            raise ValueError("No commands scheduled")
+        root, inner = self._elements
+        header, message = self._request(root)
+        ret = []
+        for message, processor in zip(message, self._processors):
+            ret.append(processor.pop(0)(message, *processor))
+        self._processors = []
+        self._elements = None
+        return ret
+
     def activate_license(self, key):
         """Activate an iLO advanced license"""
         license = etree.Element('ACTIVATE', KEY=key)
@@ -583,21 +654,21 @@ class Ilo(object):
 
     def eject_virtual_media(self, device="cdrom"):
         """Eject the virtual media attached to the specified device"""
-        return self._control_tag('RIB_INFO', 'EJECT_VIRTUAL_MEDIA', attrib={"DEVICE": device.upper()})
+        return self._control_tag('RIB_INFO', 'EJECT_VIRTUAL_MEDIA',
+                attrib={"DEVICE": device.upper()})
 
     def factory_defaults(self):
         """Reset the iLO to factory default settings"""
         return self._control_tag('RIB_INFO', 'FACTORY_DEFAULTS')
 
-    @untested
     def get_ahs_status(self):
-        """FIXME: I have no relevant hardware. Please report sample output"""
-        return self._raw(('RIB_INFO', {'MODE': 'READ'}), ('GET_AHS_STATUS', {}))
+        """Get active health system  logging status"""
+        return self._info_tag('RIB_INFO', 'GET_AHS_STATUS')
 
     def get_all_users(self):
         """Get a list of all loginnames"""
-        data = self._info_tag2('USER_INFO', 'GET_ALL_USERS', key='value')
-        return [x for x in data if x]
+        return self._info_tag2('USER_INFO', 'GET_ALL_USERS', key='value',
+                process=lambda data: [x for x in data if x])
 
     def get_all_user_info(self):
         """Get basic and authorization info of all users"""
@@ -613,30 +684,22 @@ class Ilo(object):
 
     def get_embedded_health(self):
         """Get server health information"""
-        root, attach = self._root_element('SERVER_INFO', MODE='read')
-        etree.SubElement(attach, 'GET_EMBEDDED_HEALTH')
-
-        header, message = self._request(root)
-
-        health = {}
-        for category in message.find('GET_EMBEDDED_HEALTH_DATA'):
-            tag = category.tag.lower()
-            if not list(category):
-                health[tag] = None
-            elif not list(category[0]):
-                health[tag] = {}
-                for elt in category:
-                    elttag = elt.tag.lower()
-                    if elttag not in health[tag]:
-                        health[tag][elttag] = {}
-                    health[tag][elttag].update(self._element_to_dict(elt))
-            else:
-                health[tag] = [self._element_children_to_dict(x) for x in category]
-                if 'location' in health[tag][0]:
-                    health[tag] = dict((x['location'], x) for x in health[tag])
-                elif 'label' in health[tag][0]:
-                    health[tag] = dict((x['label'], x) for x in health[tag])
-        return health
+        def process(data):
+            for category in data:
+                if category == 'health_at_a_glance':
+                    health = {}
+                    for key, val in data[category]:
+                        if key not in health:
+                            health[key] = val
+                        else:
+                            health[key].update(val)
+                    data[category] = health
+                else:
+                    tag = 'label' in data[category][0] and 'label' or 'location'
+                    data[category] = dict([(x[tag], x) for x in data[category]])
+            return data
+        return self._info_tag('SERVER_INFO', 'GET_EMBEDDED_HEALTH', 'GET_EMBEDDED_HEALTH_DATA',
+                process=process)
 
     def get_fw_version(self):
         """Get the iLO firmware version"""
@@ -651,26 +714,11 @@ class Ilo(object):
            where human readable information is available are returned. To get
            all records pass :attr:`decoded_only=False` """
 
-        root, attach = self._root_element('SERVER_INFO', MODE='read')
-        etree.SubElement(attach, 'GET_HOST_DATA')
-
-        header, message = self._request(root)
-
-        records = []
-        for record in message.find('GET_HOST_DATA').findall('SMBIOS_RECORD'):
-            if decoded_only and not list(record):
-                continue
-            record_ = self._element_to_dict(record)
-            record_['fields'] = []
-            for field in list(record):
-                record_['fields'].append(self._element_to_dict(field))
-            names = [x['name'] for x in record_['fields']]
-            if len(names) == len(set(names)):
-                for field in record_['fields']:
-                    record_[field['name']] = field['value']
-                del record_['fields']
-            records.append(record_)
-        return records
+        def process(data):
+            if decoded_only:
+                data = [x for x in data if len(x) > 2]
+            return data
+        return self._info_tag('SERVER_INFO', 'GET_HOST_DATA', process=process)
 
     def get_host_power_saver_status(self):
         """Get the configuration of the ProLiant power regulator"""
@@ -678,13 +726,13 @@ class Ilo(object):
 
     def get_host_power_status(self):
         """Whether the server is powered on or not"""
-        data = self._info_tag('SERVER_INFO', 'GET_HOST_POWER_STATUS', 'GET_HOST_POWER')
-        return data['host_power']
+        return self._info_tag('SERVER_INFO', 'GET_HOST_POWER_STATUS', 'GET_HOST_POWER',
+                process=lambda data: data['host_power'])
 
     def get_host_pwr_micro_ver(self):
         """Get the version of the power micro firmware"""
-        data = self._info_tag('SERVER_INFO', 'GET_HOST_PWR_MICRO_VER')
-        return data['pwr_micro']['version']
+        return self._info_tag('SERVER_INFO', 'GET_HOST_PWR_MICRO_VER',
+                process=lambda data: data['pwr_micro']['version'])
 
     def get_ilo_event_log(self):
         """Get the full iLO event log"""
@@ -709,10 +757,11 @@ class Ilo(object):
     def get_one_time_boot(self):
         """Get the one time boot state of the host"""
         # Inconsistency between iLO 2 and 3, let's fix that
-        data = self._info_tag('SERVER_INFO', 'GET_ONE_TIME_BOOT', ('ONE_TIME_BOOT', 'GET_ONE_TIME_BOOT'))
-        if 'device' in data['boot_type']:
-            data['boot_type'] = data['boot_type']['device']
-        return data
+        def process(data):
+            if 'device' in data['boot_type']:
+                data['boot_type'] = data['boot_type']['device']
+            return data
+        return self._info_tag('SERVER_INFO', 'GET_ONE_TIME_BOOT', ('ONE_TIME_BOOT', 'GET_ONE_TIME_BOOT'), process=process)
 
     def get_persistent_boot(self):
         """Get the boot order of the host"""
@@ -720,8 +769,7 @@ class Ilo(object):
 
     def get_power_cap(self):
         """Get the power cap setting"""
-        data = self._info_tag('SERVER_INFO', 'GET_POWER_CAP')
-        return data['power_cap']
+        return self._info_tag('SERVER_INFO', 'GET_POWER_CAP', process=lambda data: data['power_cap'])
 
     def get_power_readings(self):
         """Get current, min, max and average power readings"""
@@ -733,8 +781,7 @@ class Ilo(object):
 
     def get_server_auto_pwr(self):
         """Get the automatic power on delay setting"""
-        data = self._info_tag('SERVER_INFO', 'GET_SERVER_AUTO_PWR')
-        return data['server_auto_pwr']
+        return self._info_tag('SERVER_INFO', 'GET_SERVER_AUTO_PWR', process=lambda data: data['server_auto_pwr'])
 
     def get_server_event_log(self):
         """Get the IML log of the server"""
@@ -742,13 +789,11 @@ class Ilo(object):
 
     def get_server_name(self):
         """Get the name of the server this iLO is managing"""
-        name = self._info_tag('SERVER_INFO', 'GET_SERVER_NAME', 'SERVER_NAME')
-        return name['value']
+        return self._info_tag('SERVER_INFO', 'GET_SERVER_NAME', 'SERVER_NAME', process=lambda name: name['value'])
 
     def get_server_power_on_time(self):
         """How many minutes ago has the server been powered on"""
-        minutes = self._info_tag('SERVER_INFO', 'GET_SERVER_POWER_ON_TIME', 'SERVER_POWER_ON_MINUTES')
-        return int(minutes['value'])
+        return self._info_tag('SERVER_INFO', 'GET_SERVER_POWER_ON_TIME', 'SERVER_POWER_ON_MINUTES', process=lambda data: int(data['value']))
 
     def get_snmp_im_settings(self):
         """Where does the iLO send SNMP traps to and which traps does it send"""
@@ -756,22 +801,7 @@ class Ilo(object):
 
     def get_sso_settings(self):
         """Get the HP SIM Single Sign-On settings"""
-        root, attach = self._root_element('SSO_INFO', MODE='read')
-        etree.SubElement(attach, 'GET_SSO_SETTINGS')
-
-        header, message = self._request(root)
-
-        retval = {}
-        for record in message.find('GET_SSO_SETTINGS'):
-            tag = record.tag.lower()
-            attrib = record.attrib
-            if 'VALUE' in attrib:
-                retval[tag] = self._coerce(attrib['VALUE'])
-                continue
-            if tag not in retval:
-                retval[tag] = {}
-            retval[tag].update(dict([(x[0].lower(), self._coerce(x[1])) for x in attrib.items()]))
-        return retval
+        return self._info_tag('SSO_INFO', 'GET_SSO_SETTINGS')
 
     def get_twofactor_settings(self):
         """Get two-factor authentication settings"""
@@ -779,8 +809,7 @@ class Ilo(object):
 
     def get_uid_status(self):
         """Get the status of the UID light"""
-        data = self._info_tag('SERVER_INFO', 'GET_UID_STATUS')
-        return data['uid']
+        return self._info_tag('SERVER_INFO', 'GET_UID_STATUS', process=lambda data: data['uid'])
 
     def get_user(self, user_login):
         """Get user info about a specific user"""
@@ -869,6 +898,50 @@ class Ilo(object):
                     for x in vars if vars[x] is not None]
         return self._control_tag('RIB_INFO', 'MOD_NETWORK_SETTINGS', elements=elements)
 
+
+    def mod_dir_config(self, dir_authentication_enabled=None,
+            dir_local_user_acct=None,dir_server_address=None,
+            dir_server_port=None,dir_object_dn=None,dir_object_password=None,
+            dir_user_context_1=None,dir_user_context_2=None,
+            dir_user_context_3=None,dir_user_context_4=None,
+            dir_user_context_5=None,dir_user_context_6=None,
+            dir_user_context_7=None,dir_user_context_8=None,
+            dir_user_context_9=None,dir_user_context_10=None,
+            dir_user_context_11=None,dir_user_context_12=None,
+            dir_user_context_13=None,dir_user_context_14=None,
+            dir_user_context_15=None,dir_enable_grp_acct=None,
+            dir_kerberos_enabled=None,dir_kerberos_realm=None,
+            dir_kerberos_kdc_address=None,dir_kerberos_kdc_port=None,
+            dir_kerberos_keytab=None,
+            dir_grpacct1_name=None,dir_grpacct1_sid=None,
+            dir_grpacct1_priv=None,dir_grpacct2_name=None,
+            dir_grpacct2_sid=None,dir_grpacct2_priv=None,
+            dir_grpacct3_name=None,dir_grpacct3_sid=None,
+            dir_grpacct3_priv=None,dir_grpacct4_name=None,
+            dir_grpacct4_sid=None,dir_grpacct4_priv=None,
+            dir_grpacct5_name=None,dir_grpacct5_sid=None,
+            dir_grpacct5_priv=None,dir_grpacct6_name=None,
+            dir_grpacct6_sid=None,dir_grpacct6_priv=None):
+        """Modify iLO directory configuration, only values that are specified
+             will be changed."""
+        vars = dict(locals())
+        del vars['self']
+
+        # create special case for element with text inside
+        if dir_kerberos_keytab:
+            keytab_el = etree.Element('DIR_KERBEROS_KEYTAB')
+            keytab_el.text = dir_kerberos_keytab
+            del vars['dir_kerberos_keytab']
+
+        elements = [etree.Element(x.upper(), VALUE=str({True: 'Yes', \
+                False: 'No'}.get(vars[x], vars[x])))
+                    for x in vars if vars[x] is not None]
+
+        if dir_kerberos_keytab:
+            elements.append(keytab_el)
+        return self._control_tag('DIR_INFO','MOD_DIR_CONFIG',elements=elements)
+
+
     def mod_snmp_im_settings(self, web_agent_ip_address=None, snmp_address_1=None,
             snmp_address_2=None, snmp_address_3=None, os_traps=None,
             snmp_passthrough_status=None, rib_traps=None, cim_security_mask=None):
@@ -921,10 +994,9 @@ class Ilo(object):
         """Power cycle the server"""
         return self._control_tag('SERVER_INFO', 'RESET_SERVER')
 
-    @untested
     def set_ahs_status(self, status):
         """Enable or disable AHS logging"""
-        status = {True: 'Enable', False: 'Disable'}[status]
+        status = {True: 'enable', False: 'disable'}[status]
         return self._control_tag('RIB_INFO', 'SET_AHS_STATUS', attrib={'VALUE': status})
 
     def set_language(self, lang_id):
@@ -1004,6 +1076,9 @@ class Ilo(object):
     def update_rib_firmware(self, filename, progress=None):
         """Upload new RIB firmware, use "latest" as filename to automatically
         download and use the latest firmware"""
+        if self.delayed:
+            raise IloError("Cannot run firmware update in delayed mode")
+
         if not self.protocol:
             self._detect_protocol()
 
